#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
GNUv3 Licensed.
Contact at www.sinclair.bio
"""

# Use the optmagic library to make a command line tool automatically #
from optmagic import OptMagic

# The main object of our package #
from crest4 import Classify

# The main function to run when we are called #
def main():
    magic = OptMagic(Classify)
    return magic()

# Execute when run, not when imported #
if __name__ == "__main__": main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script to run the `vsearch` integration test.
"""

# Built-in modules #
import inspect

# First party modules #
from autopaths import Path

# Third party modules #

# Internal modules #
from crest4 import Classify

# Get the current directory of this python script #
this_file = Path((inspect.stack()[0])[1])
this_dir  = this_file.directory

###############################################################################
def test_vsearch():
    # The input fasta #
    fasta = this_dir.find('*.fasta')
    # The output directory #
    output_dir = this_dir + 'results/'
    output_dir.remove()
    # Create object #
    c = Classify(fasta       = fasta,
                 output_dir  = output_dir,
                 search_algo = 'vsearch',
                 num_threads = True)
    # Run it #
    c()
    # Check that the results are good #
    assert c.queries_by_id['Kocuria'].taxonomy[-10] == "Kocuria"
    assert c.queries_by_id['Marmoricola'].taxonomy[-10] == "Marmoricola"
    # Return #
    return c

###############################################################################
if __name__ == '__main__':
    classify = test_vsearch()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
GNUv3 Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import os, json

# Internal modules #
import crest4

# First party modules #
from autopaths.dir_path  import DirectoryPath
from autopaths.file_path import FilePath
from plumbing.cache      import property_cached
from plumbing.scraping   import download_from_url, retrieve_from_url

###############################################################################
class CrestMetadata:
    """
    The databases that `crest4` needs are slightly too large to be distributed
    with the source tarball on PyPI. Therefore, we host them on a separate
    server. Currently, this is S3 on AWS.

    Over the course of years, however, these servers might be shutdown.
    To avoid this inconvenience, instead of hard-coding the server address here
    in the source file, we instead maintain a metadata file.

    This metadata file contains the up-to-date URLs of the databases
    to download. It is placed in the GitHub repository, and only the location
    of the metadata file is hardcoded into the source code.

    This enables updating the URLs without having to issue a new release of
    `crest4`.

    A good read on this subject is the following:

    https://www.dampfkraft.com/code/distributing-large-files-with-pypi.html
    """

    # The location of the metadata file is hardcoded #
    base_url  = "https://raw.githubusercontent.com/xapple/"
    repo_path = "crest4/master/crest4/crest4_db_urls.json"

    # This will always retrieve the raw content of the latest version #
    metadata_url = base_url + repo_path

    # To see the HTML version on GitHub, simply head to this address instead:
    # https://github.com/xapple/crest4/blob/master/crest4/crest4_db_urls.json

    @property_cached
    def db_urls(self):
        """
        Once the `Metadata` object is created simply access it like this:
        >>> print(metadata.db_urls['silvamod138']['url'])
        """
        # Download the gist #
        content = retrieve_from_url(self.metadata_url)
        # Parse it #
        data = json.loads(content)
        # Validate it #
        assert list(data) == ['crest4_databases']
        # Return it #
        return data['crest4_databases']

# The metadata should very rarely be updated -- we create a singleton #
metadata = CrestMetadata()

###############################################################################
class CrestDatabase:
    """
    This object represents a database of taxonomically curated DNA sequences
    that was subsequently specifically formatted for use with `crest4`.
    It comes with a FASTA file, as well as a tree.
    """

    # Default location for a database #
    default_dir = os.path.expanduser('~/.crest4/')

    # The environment variable that the user can set #
    environ_var = "CREST4_DIR"

    def __init__(self, *args, **kwargs):
        """
        Either take one of the built-in databases by specifying:
        * name
        * desc

        Or bring your own database by specifying:
        * custom_path

        The user can specify a path pointing to a directory that contains all
        the required files.
        It must include the database FASTA file as well as the `.map` and
        `.names` file.
        Here if the user indicates a file, then take the parent directory.
        If the user has selected a directory, then use it directly and assume
        the file has the same name as the directory.
        """
        # Case where we pick a built-in database #
        if 'name' in kwargs and 'desc' in kwargs:
            # The three attributes #
            self.dir_name  = kwargs['name']
            self.file_name = kwargs['name']
            self.desc      = kwargs['desc']
            # If the user doesn't specify a base_dir path, then we should
            # determine where the actual files will be located on disk and
            # check if the user has set the special environment variable.
            if 'base_dir' not in kwargs:
                base_dir = os.environ.get(self.environ_var, self.default_dir)
            else:
                base_dir = kwargs['base_dir']
            # Convert to an absolute DirectoryPath #
            self.base_dir = DirectoryPath(base_dir)
            self.base_dir = DirectoryPath(self.base_dir.absolute_path)
        # Case where the user specifies a custom path #
        elif 'custom_path' in kwargs:
            # If the user specifies a path, then no need to download #
            self.downloaded = True
            # Check if it is a file or if it is a directory #
            custom_path = kwargs['custom_path']
            if os.path.isfile(custom_path):
                custom_path = FilePath(custom_path)
                self.base_dir  = custom_path.directory.directory
                self.dir_name  = custom_path.directory.name
                self.file_name = custom_path.prefix
            else:
                custom_path = DirectoryPath(custom_path)
                self.base_dir  = custom_path.directory
                self.dir_name  = custom_path.name
                self.file_name = custom_path.name
            # Set the description #
            self.desc = "Custom user-provided database '%s'." % self.dir_name

    def __repr__(self):
        """A simple representation of this object to avoid memory addresses."""
        return "<%s object at '%s'>" % (self.__class__.__name__, self.path)

    @property
    def tag(self):
        """
        Some methods will look for the `tag` attribute to print the name of the
        database.
        """
        return self.dir_name

    @property
    def tarball(self):
        """Determine where the database `.tar.gz` will be located on disk."""
        return self.base_dir + self.dir_name + '.tar.gz'

    @property
    def path(self):
        """The path to the FASTA file."""
        return self.base_dir + self.dir_name + '/' + self.file_name + '.fasta'

    @property_cached
    def downloaded(self):
        """
        Determine if the database has been downloaded to disk and uncompressed
        previously. Returns `True` or `False`.
        """
        # Check if the tree is there and not empty #
        return bool(self.path.replace_extension('tre'))

    @property
    def url(self):
        """
        Retrieve the URL of the database file to download by first downloading
        a metadata file (see the `Metadata` class above).
        """
        # Check that the key exists #
        if self.dir_name not in metadata.db_urls:
            msg = "The database '%s' is not in the list of downloadable" \
                  " databases. Please update the file `crest4_db_urls.json`" \
                  " to include it."
            raise Exception(msg % self.dir_name)
        # Get the URL #
        return metadata.db_urls[self.dir_name]['url']

    def download(self):
        """
        Download the database file, uncompress it, and save it to disk.
        """
        # Compose a message so the user knows why it's taking time #
        message = "The database '%s' has not been downloaded yet." \
                  " This process will start now and might take some time" \
                  " depending on your internet connection. Please be" \
                  " patient. The result will be saved to '%s'. You can" \
                  " override this by setting the $%s environment variable."
        message = message % (self.dir_name, self.base_dir, self.environ_var)
        # Display the message with style in a box #
        from plumbing.common import rich_panel_print
        rich_panel_print(message, "Large Download")
        # Show a progress bar a bit like wget #
        download_from_url(self.url,
                          destination = self.tarball,
                          uncompress  = False,
                          user_agent  = "crest4 v" + crest4.__version__,
                          stream      = True,
                          progress    = True,
                          desc        = self.dir_name)
        # Uncompress #
        self.tarball.untargz_to(self.base_dir)
        # Remove tarball #
        self.tarball.remove()
        # Remove macOS attribute files that get bundled in the tar #
        for path in self.base_dir.flat_files:
            if path.name.startswith("._"): path.remove()
        # Check it worked #
        del self.downloaded
        assert self.downloaded

    #--------------------------- Specific Indexes ----------------------------#
    @property_cached
    def blast_db(self):
        """
        Return a `BLASTDb` object that can be used for the sequence
        similarity search.
        """
        # Download the database if it has not been done already #
        if not self.downloaded: self.download()
        # Create the database object #
        from seqsearch.search.blast import BLASTdb
        db = BLASTdb(self.path, seq_type='nucl')
        # Create the database with `mkblastdb` if it's not made already #
        db.create_if_not_exists(verbose=True)
        # Return #
        return db

    @property_cached
    def vsearch_db(self):
        """
        Return a `VSEARCHdb` object that can be used for the sequence
        similarity search.
        """
        # Download the database if it has not been done already #
        if not self.downloaded: self.download()
        # Create the database object #
        from seqsearch.search.vsearch import VSEARCHdb
        db = VSEARCHdb(self.path.replace_extension('udb'))
        # Create the database with `vsearch` if it's not made already #
        db.create_if_not_exists(verbose=True)
        # Return #
        return db

    #--------------------------- Loading the tree ----------------------------#
    @property_cached
    def tree(self):
        """
        Using the `.tre` file, we return an N-ary tree in memory.
        A number characterizes every node. For instance, between 1 and 32'477.
        No other information is present in the nodes.
        """
        from ete4 import Tree
        return Tree(str(self.path.replace_extension('tre')), parser=8)

    @property_cached
    def node_to_name(self):
        """
        Using the `.names` file, we return a dictionary linking node numbers to
        a tuple of species names and a minimum similarity fraction like 0.97.
        Example: 51 -> (Proteobacteria, 0.8)
        """
        # Get the path of the file #
        path = self.path.replace_extension('names')
        # Define how to process each line #
        def parse_lines(lines):
            for line in lines:
                if line.startswith('#') or not line: continue
                num, name, frac = line.strip().split(',')
                yield num, (name, frac)
        # Create a dictionary #
        with open(path, 'rt') as handle: return dict(parse_lines(handle))

    @property_cached
    def acc_to_node(self):
        """
        Using the `.map` file, we return a dictionary linking accession
        strings (found in the `.fasta` file for each sequence) to node numbers.
        Example: HQ191339 -> 28386
        """
        # Get the path of the file #
        path = self.path.replace_extension('map')
        # Check that it exists #
        if not path.exists:
            msg = ("The file '%s' does not exist. "
                   "Contents of the parent directory:\n%s")
            contents = '- ' + '\n- '.join(path.directory.flat_files)
            raise FileNotFoundError(msg % (path, contents))
        # Define how to process each line #
        def parse_lines(lines):
            for line in lines:
                if line.startswith('#') or not line: continue
                num, name = line.strip().split(',')
                yield name, num
        # Create a dictionary #
        with open(path, 'rt') as handle: return dict(parse_lines(handle))

    #----------------------------- Sanity checks -----------------------------#
    # See file `analyze_tre_files.py` in the `crest4_utils` repository.
    pass

    #--------------------------- Extra information ---------------------------#
    @property
    def rank_names(self):
        return ['Root',           # 0 (This is life itself, called 'meta' here)
                'Genome',         # 1 (e.g. 'Plastid' or 'Main Genome')
                'Domain',         # 2 (This is Bacteria, Archaea or Eucarya)
                'Superkingdom',   # 3 (e.g. 'Bacteria (superkingdom)')
                'Kingdom',        # 4 (This is also called <Superphylum>)
                'Phylum',         # 5 (e.g. 'Proteobacteria' or 'Firmicutes')
                'Class',          # 6  (e.g. 'Bacilli')
                'Order',          # 7  (e.g. 'Bacillales')
                'Family',         # 8  (e.g. 'Bacillaceae')
                'Genus',          # 9  (e.g. 'Bacillus')
                'Species',        # 10 (e.g. 'Bacillus subtilis')
                'Strain']         # 11 (e.g. 'Bacillus subtilis 168')

###############################################################################
# As our databases should only be stored on disk once, so we have singletons #
midori253darn = CrestDatabase(name = 'midori253darn',
                              desc = 'The midori253darn database for crest4')

mitofish = CrestDatabase(name = 'mitofish',
                         desc = 'MitoFish')

silvamod138pr2 = CrestDatabase(name = 'silvamod138pr2',
                               desc = 'Silva version 138 PR2 version 2')

ssuome = CrestDatabase(name = 'ssuome',
                       desc = 'Combination of 16S sequences from SILVA(mod)'
                              ' SSURef v138 and the Eukaryome database'
                              ' (v1.9.3)')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
The `crest4` python package can automatically assign taxonomic names to
DNA sequences obtained from environmental sequencing.

The README and source code are located at:

https://github.com/xapple/crest4
"""

# Special variables #
__version__ = '4.4.5'

# Constants #
project_url = 'https://github.com/xapple/crest4'

# Expose our main object at the module level
# So that you can just do `from crest4 import Classify` later
from .classify import Classify



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script to run the `cmd_line_tool` integration test.
"""

# Built-in modules #
import inspect, sys, subprocess

# First party modules #
from autopaths import Path

# Third party modules #
import pytest, asyncio

# Internal modules #

# Get the current directory of this python script #
this_file = Path((inspect.stack()[0])[1])
this_dir  = this_file.directory

###############################################################################
@pytest.mark.asyncio
async def test_cmd_line_tool():
    # The input fasta #
    fasta = this_dir.find('*.fasta')
    # The output directory #
    output_dir = this_dir + 'results/'
    output_dir.remove()
    # Build the command as a list of arguments
    cmd = [sys.executable,
           '-m', 'crest4',
           '--fasta', fasta,
           '--output_dir', output_dir,
           '--num_threads', 'True']
    # Run the command asynchronously
    process = await asyncio.create_subprocess_exec(*cmd,
                                                   stdout = subprocess.PIPE,
                                                   stderr = subprocess.PIPE)
    # Wait for the process to finish #
    stdout, stderr = await process.communicate()
    # Check that the results were created #
    created_file = output_dir + 'assignments.txt'
    assert created_file
    # Check that the return code is 0 #
    assert process.returncode == 0
    # Return #
    return process

###############################################################################
if __name__ == '__main__':
    proc = test_cmd_line_tool()
    asyncio.run(proc)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script to run the `midori253darn` database test.
"""

# Built-in modules #
import inspect, os

# First party modules #
from autopaths import Path

# Third party modules #
import pytest

# Internal modules #
from crest4 import Classify

# Get the current directory of this python script #
this_file = Path((inspect.stack()[0])[1])
this_dir  = this_file.directory

###############################################################################
@pytest.mark.skipif(
    os.getenv("GITHUB_ACTIONS") == "true",
    reason="GitHub limits cache to 10GB, this DB is too big.",
)
def test_db_midori():
    # The input fasta #
    fasta = this_dir.find('*.fasta')
    # The output directory #
    output_dir = this_dir + 'results/'
    output_dir.remove()
    # Create object #
    c = Classify(fasta       = fasta,
                 output_dir  = output_dir,
                 search_algo = 'vsearch',
                 search_db   = 'midori253darn',
                 num_threads = True)
    # Run it #
    c()
    # Check that the results are good #
    assert c.queries_by_id['Test'].taxonomy[0] == "Saccharomyces"
    # Return #
    return c

###############################################################################
if __name__ == '__main__':
    classify = test_db_midori()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script to run the `with_otu_table` integration test.
"""

# Built-in modules #
import inspect

# First party modules #
from autopaths import Path

# Third party modules #
import pytest

# Internal modules #
from crest4 import Classify

# Get the current directory of this python script #
this_file = Path((inspect.stack()[0])[1])
this_dir  = this_file.directory

###############################################################################
def test_with_bad_otu_table():
    # The input fasta #
    fasta = this_dir.find('*.fasta')
    # The input table #
    otu_table = this_dir.find('*bad.tsv')
    # The output directory #
    output_dir = this_dir + 'results_bad/'
    output_dir.remove()
    # Create object #
    c = Classify(fasta       = fasta,
                 output_dir  = output_dir,
                 num_threads = True,
                 otu_table   = otu_table)
    # It should complain when we run it #
    with pytest.raises(ValueError): c()

###############################################################################
def test_with_good_otu_table():
    # The input fasta #
    fasta = this_dir.find('*.fasta')
    # The input table #
    otu_table = this_dir.find('*good.tsv')
    # The output directory #
    output_dir = this_dir + 'results_good/'
    output_dir.remove()
    # Create object #
    c = Classify(fasta       = fasta,
                 output_dir  = output_dir,
                 num_threads = True,
                 otu_table   = otu_table)
    # Run it #
    c()
    # Check that the results are good #
    assert all(c.otu_info.otus_by_rank['river'].array == [51, 10])
    assert all(c.otu_info.otus_by_rank['lake'].array  == [102, 50])
    assert all(c.otu_info.otus_by_rank['ocean'].array == [3, 200])
    # Return #
    return c

###############################################################################
if __name__ == '__main__':
    # The bad case #
    classify_bad = test_with_bad_otu_table()
    # The good case #
    classify_good = test_with_good_otu_table()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script to run the `score_drop_change` integration test.
"""

# Built-in modules #
import inspect, subprocess, sys

# First party modules #
from autopaths import Path

# Internal modules #

# Get the current directory of this python script #
this_file = Path((inspect.stack()[0])[1])
this_dir  = this_file.directory

###############################################################################
def test_score_drop_change():
    # The input fasta #
    fasta = this_dir.find('*.fasta')
    # The output directory #
    output_dir = this_dir + 'results/'
    output_dir.remove()
    # Call via the command line tool #
    result = subprocess.run([sys.executable,
                             '-m',            'crest4',
                             '--fasta',       str(fasta),
                             '--output_dir',  str(output_dir),
                             '--score_drop',  '1'],
                            check=True)
    # Check that the results were created #
    created_file = output_dir + 'assignments.txt'
    assert created_file
    # Return #
    return result

###############################################################################
if __name__ == '__main__':
    test_score_drop_change()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
GNUv3 Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import os, multiprocessing

# Internal modules #
import crest4
import crest4.databases
from crest4.query import Query
from crest4.databases import CrestDatabase

# First party modules #
from functools import cached_property
from autopaths.file_path import FilePath
from autopaths.dir_path  import DirectoryPath

# Constants #
all_db_choices = ('midori253darn', 'silvamod138pr2', 'mitofish', 'ssuome')

###############################################################################
class Classify:
    """
    This is the main object offered by the `crest4` package.
    It enables you to automatically assign taxonomic names to DNA sequences
    obtained from environmental sequencing.

    After creating a new instance of a `Classify` object, you can simply call
    it to get your input data processed and the assignments output file
    generated. Examples are included in the `README.md` file of this package,
    which can directly be seen on the GitHub page at:

    https://github.com/xapple/crest4/
    """

    def __init__(self,
                 fasta,
                 search_algo = 'blast',
                 num_threads = 1,
                 search_db   = 'ssuome',
                 output_dir  = None,
                 search_hits = None,
                 min_score   = None,
                 score_drop  = 2.0,
                 min_smlrty  = True,
                 otu_table   = None,
                 ):
        """
        Args:

            fasta: The path to a single FASTA file as a string.
                   These are the sequences that will be taxonomically
                   classified.

            search_algo: The algorithm used for the sequence similarity search
                         that will be run to match the sequences against the
                         database chosen. Either `blast` or `vsearch`. No
                         other values are currently supported. By default,
                         `blast`.

            num_threads: The number of processors to use for the sequence
                         similarity search. By default, parallelism is turned
                         off, and this value is 1. If you pass the value
                         `True,` we will run as many processes as there are
                         CPUs but no more than 32.

            search_db: The database used for the sequence similarity search.
                       Either `ssuome`, `silvamod138pr2`, 'mitofish',
                       or `midori253darn`. By default, `ssuome`.
                       Optionally, the user can provide a custom database by
                       specifying the full path to a directory containing all
                       required files under `search_db`.
                       See the README for more information.

            output_dir: The directory into which all the classification
                        results will be written to. This defaults to a
                        directory with the same name as the original FASTA
                        file and a `.crest4` suffix appended.

            search_hits: The path where the search results will be stored.
                         This defaults to the output directory. However,
                         if the search operation has already been completed
                         beforehand, specify the path here to skip the
                         sequence similarity search step and go directly to
                         the taxonomy step. If a hits file exists in the output
                         directory and this option is not specified, it is
                         deleted and regenerated.

            min_score: The minimum bit-score for a search hit to be considered
                       when using BLAST as the search algorithm. All hits below
                       this score are ignored. When using VSEARCH, this value
                       instead indicates the minimum identity between two
                       sequences for the hit to be considered.
                       The default is `155` for BLAST and `0.75` for VSEARCH.

            score_drop: Determines the range of hits to retain and the range
                        to discard based on a drop in percentage from the score
                        of the best hit. Any hit below the following value:
                        "(100 - score_drop)/100 * best_hit_score" is ignored.
                        By default, `2.0`.

            min_smlrty: Determines if the minimum similarity filter is turned
                        on or off. Pass any value like `False` to turn it off.
                        The minimum similarity filter prevents classification
                        to higher ranks when a minimum rank-identity is not
                        met. The default is `True`.

            otu_table: Optionally, one can specify the path to an existing OTU
                       table in CSV or TSV format when running `crest4`.
                       The sequence names in the OTU table must be rows and
                       have to match the names in the FASTA file. The column,
                       on the other hand, provide your samples names.
                       When this option is used, then two extra output files
                       are generated. Firstly, a table summarizing the
                       assignment counts per taxa. Secondly, a table
                       propagating the sequence counts upwards
                       in a cumulative fashion.
                       """
        # Save attributes #
        self.fasta       = fasta
        self.search_algo = search_algo
        self.num_threads = num_threads
        self.search_db   = search_db
        self.output_dir  = output_dir
        self.search_hits = search_hits
        self.min_score   = min_score
        self.score_drop  = score_drop
        self.min_smlrty  = min_smlrty
        self.otu_table   = otu_table
        # Assign default values and change others #
        self.transform()
        # Validate attributes #
        self.validate()

    def transform(self):
        """
        This method will replace empty attributes with defaults when this is
        needed and will convert others to proper types.
        """
        # The fasta should be a FASTA object #
        if self.fasta is not None:
            from fasta import FASTA
            self.fasta = FASTA(self.fasta)
        # Default for the number of threads #
        if not isinstance(self.num_threads, int):
            if self.num_threads is True:
                self.num_threads = min(multiprocessing.cpu_count(), 32)
            elif self.num_threads is False:
                self.num_threads = 1
            elif self.num_threads.lower() == 'true':
                self.num_threads = min(multiprocessing.cpu_count(), 32)
        # The database is always a string #
        self.search_db = str(self.search_db)
        # Default for the output directory #
        if self.output_dir is None:
            self.output_dir = self.fasta + '.crest4/'
        self.output_dir = DirectoryPath(self.output_dir)
        # The search hits is a file somewhere if passed #
        if self.search_hits is not None:
            self.search_hits = FilePath(self.search_hits)
        # Default for the search hits file if not passed #
        if self.search_hits is None:
            self.search_hits = FilePath(self.output_dir + 'search.hits')
            self.search_hits.remove()
        # Default for the minimum score #
        if self.min_score is None:
            if self.search_algo == 'blast':
                self.min_score = 155
            if self.search_algo == 'vsearch':
                self.min_score = 0.75
        # The minimum score has to be a number, not a string #
        try:
            self.min_score = float(self.min_score)
        except (ValueError, TypeError):
            msg = "The minimum score value must be numerical (not '%s')."
            raise ValueError(msg % self.min_score)
        # The score drop has to be a float, not a string #
        try:
            self.score_drop = float(self.score_drop)
        except (ValueError, TypeError):
            msg = "The score drop value must be numerical (not '%s')."
            raise ValueError(msg % self.score_drop)
        # Turn off the minimum similarity filter if the user passed any value #
        if self.min_smlrty is not True: self.min_smlrty = False
        # The OTU table is a file somewhere if passed #
        if self.otu_table is not None:
            self.otu_table = FilePath(self.otu_table)
            self.otu_table.must_exist()

    def validate(self):
        """
        This method will raise an Exception if any of the arguments passed by
        the user are illegal.
        """
        # The fasta should exist if passed #
        if self.fasta is not None:
            self.fasta.must_exist()
        # Either the FASTA file or the hits file has to contain something #
        if not self.fasta and not self.search_hits:
            msg = "Neither the FASTA file at '%s' nor the search hits file" \
                  " at '%s' contain any data. Cannot proceed."
            raise Exception(msg % (self.fasta, self.search_hits))
        # Check the search algorithm #
        if self.search_algo not in ('blast', 'vsearch'):
            msg = "The search algorithm '%s' is not supported."
            raise ValueError(msg % self.search_algo)
        # The search database is a known entry or exists on the filesystem #
        if self.search_db not in all_db_choices:
            if not os.path.exists(self.search_db):
                msg = "The search database '%s' is not supported."
                raise ValueError(msg % self.search_db)
        # Check the minimum score value is above zero #
        if self.min_score < 0.0:
            msg = "The minimum score cannot be smaller than zero ('%s')."
            raise ValueError(msg % self.min_score)
        # Check the minimum score value is below one #
        if self.min_score > 1.0:
            if self.search_algo == 'vsearch':
                msg = "The minimum score cannot be more than 1.0 when" \
                      " using VSEARCH ('%s') because it represents the" \
                      " the minimum identity between two sequences."
                raise ValueError(msg % self.min_score)
        # Check the score drop value #
        if self.score_drop < 0.0:
            msg = "The score drop value cannot be smaller than zero ('%s')."
            raise ValueError(msg % self.min_score)
        if self.score_drop > 100.0:
            msg = "The score drop value cannot be over 100 ('%s')."
            raise ValueError(msg % self.min_score)

    def __repr__(self):
        """A simple representation of this object to avoid memory addresses."""
        return "<%s object on '%s'>" % (self.__class__.__name__, self.fasta)

    @cached_property
    def database(self):
        """
        Retrieve the database object that the user has selected.
        This can be either a standard database or a custom-specified path.
        """
        if self.search_db not in all_db_choices:
            # Take the absolute path #
            user_path = os.path.abspath(self.search_db)
            # Make the object #
            return CrestDatabase(custom_path=user_path)
        else:
            return getattr(crest4.databases, self.search_db)

    #------------------------------ Searching --------------------------------#
    @cached_property
    def seqsearch(self):
        """
        An object representing the sequence similarity search.
        Makes use of the `seqsearch` module. For reference:

        * Setting `-outfmt` to 5 means XML output.
        * Setting `-outfmt` to 6 means tabular output.
        * Setting `-outfmt` to 7 means tabular output with comments.
        """
        # Initialize #
        params = {}
        # If the user chose BLAST then we have to specify tabular output #
        if self.search_algo == 'blast':
            params = {'-outfmt': '7 qseqid sseqid bitscore length nident'}
        # In case the user chose VSEARCH we specify the minimum identity
        # and the minimum sequence match length
        if self.search_algo == 'vsearch':
            params = {'--id':      self.min_score,
                      '--mincols': 25}
        # Build and return the object #
        from seqsearch.search import SeqSearch
        return SeqSearch(input_fasta = self.fasta,
                         database    = self.database,
                         seq_type    = 'nucl',
                         algorithm   = self.search_algo,
                         filtering   = {'max_targets': 100},
                         num_threads = self.num_threads,
                         out_path    = self.search_hits,
                         params      = params)

    def search(self):
        """A method to launch the sequence similarity search."""
        # Launch the search algorithm #
        return self.seqsearch.run()

    #----------------------------- Assigning ---------------------------------#
    @cached_property
    def score_frac(self):
        """
        Using the parameter `self.score_drop`, which is a percentage (e.g., 2)
        indicating a drop, we compute the minimum remaining amount of score
        allowed, as a fraction (e.g., 0.98).
        """
        return 1 - (self.score_drop / 100)

    @cached_property
    def queries(self):
        """
        Parses the output of the sequence search program used which returns
        a list of `QueryResult` objects from the `SearchIO.parse` module.
        The list contains one Query object per sequence that was originally
        inputted. Use these objects to access the taxonomic assignments.
        """
        # Check if the search has been done already #
        if not self.search_hits: self.search()
        # Iterate on the sequence search results #
        result = [Query(self, query) for query in self.seqsearch.results]
        # VSEARCH entirely forgets about sequences that had no hits.
        # Instead of still listing them in the output like BLAST.
        # So we have to add them back to the list in this awkward manner
        if self.search_algo == 'vsearch':
            reported_names = set(query.name for query in result)
            for seq in self.fasta:
                if seq.id not in reported_names:
                    q = type('FakeQuery', (), {'hits': [], 'id': seq.id})
                    result.append(Query(self, q))
        # Return #
        return result

    @cached_property
    def queries_by_id(self):
        """
        References the same Query objects as the `queries` property above,
        except that this time they are in a dictionary with the query ids
        (i.e., the original fasta ids) as keys instead of in a list.
        """
        return {query.name: query for query in self.queries}

    #------------------------------- Outputs ---------------------------------#
    @cached_property
    def out_file(self):
        """
        The path to the file that will contain the taxonomic assignments
        for every sequence.
        """
        # Make sure that the output directory exists #
        self.output_dir.create_if_not_exists()
        # Return #
        return self.output_dir + "assignments.txt"

    @cached_property
    def otu_info(self):
        """An object giving access to the OTU table information and methods."""
        from crest4.otu_tables import InfoFromTableOTUs
        return InfoFromTableOTUs(self, self.otu_table)

    def __call__(self):
        """Generate outputs."""
        # Intro message #
        print('Running crest4 version ' + crest4.__version__)
        # Iterate #
        self.out_file.writelines(query.tax_string for query in self.queries)
        # Special case where an OTU table was passed #
        if self.otu_table:
            path_by_rank    = self.output_dir + 'otus_by_rank.tsv'
            path_cumulative = self.output_dir + 'otus_cumulative.tsv'
            otus_by_rank    = self.otu_info.otus_by_rank
            otus_cumulative = self.otu_info.otus_cumulative
            otus_by_rank.to_csv(path_by_rank, index=False, sep='\t')
            otus_cumulative.to_csv(path_cumulative, index=False, sep='\t')
        # Print a success message #
        msg = "Classification ran successfully. Results are placed in '%s'."
        print(msg % self.out_file)
        # Return #
        return self.out_file


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
GNUv3 Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import inspect

# First party modules #
from seqsearch.databases.download import acc_to_fasta
from fasta import FASTA
from autopaths import Path

# Third party modules #
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq

# Get the current directory of this python script #
this_file = Path((inspect.stack()[0])[1])
this_dir  = this_file.directory

# Random sequences from NCBI #
# Taken from https://www.ncbi.nlm.nih.gov/refseq/targetedloci/16S_process/ #
accessions = {
    'NR_171545.1': "Bacteria; Actinobacteria; Micrococcales; Micrococcaceae; Kocuria",
    'NR_171544.1': "Bacteria; Actinobacteria; Propionibacteriales; Nocardioidaceae; Marmoricola"
}

###############################################################################
if __name__ == '__main__':
    # Download from the web #
    sequences = acc_to_fasta(accessions.keys())

    # Parse them #
    records = [(seq['TSeq_sequence'],
                ' '.join(seq['TSeq_defline'].split(' ')[:3]))
               for seq in sequences]

    # Make biopython objects out of them #
    records = [SeqRecord(Seq(seq), id=name, description='')
               for seq, name in records]

    # Write to a new file #
    fasta = FASTA(this_dir + 'two_seqs.fasta')
    fasta.write(records)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
GNUv3 Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
from collections import defaultdict

# First party modules #
from plumbing.cache import property_cached

###############################################################################
class InfoFromTableOTUs:
    """
    This class accepts a path to an OTU table as input.
    It will parse the table, along with the taxonomic assignments
    from the `Classify` object, and it will produce two extra output files.
    """

    def __init__(self, classify, otu_table):
        # A reference to the parent object #
        self.classify = classify
        # Path to a CSV or TSV file #
        self.otu_table = otu_table

    def __repr__(self):
        """A simple representation of this object to avoid memory addresses."""
        return "<%s object on '%s'>" % (self.__class__.__name__, self.otu_table)

    @property_cached
    def format(self):
        """
        Automatically detect the format between "TSV" or "CSV" and return the
        appropriate separator character. By default, we will return a tab.
        """
        if self.otu_table.filename.split('.')[-1] == 'csv':
            return ','
        return '\t'

    @property_cached
    def otus_df(self):
        """Load the otu_table file as a pandas `DataFrame`."""
        # Load from a text file #
        import pandas
        df = pandas.read_csv(str(self.otu_table), sep=self.format, index_col=0)
        # We only want the very first part of the IDs #
        df.index = df.index.map(lambda s: s.split()[0])
        # Return #
        return df

    @property_cached
    def otus_by_rank(self):
        """The first output file where cumulativeness is turned off."""
        return self(cumulative=False)

    @property_cached
    def otus_cumulative(self):
        """The second output file where cumulativeness is turned on."""
        return self(cumulative=True)

    def check_id_match(self):
        """
        Check that all the IDs in the OTU table given by the user match the
        ones in the FASTA file.
        """
        for i, otu_counts in self.otus_df.iterrows():
            # The name or id of the current OTU #
            name = otu_counts.name
            if name not in self.classify.queries_by_id:
                msg = "The sequence named '%s' in the table located at '%s'" \
                      " does not appear in the hits file provided at '%s'." \
                      " This can be due to the original FASTA file not" \
                      " containing them either, or because the sequence" \
                      " search was interrupted and a partial output created."
                msg = msg % (name, self.otu_table, self.classify.search_hits)
                raise ValueError(msg)

    def __call__(self, cumulative=False):
        # Let's first check all the IDs are found #
        self.check_id_match()
        # An empty pandas Series with each sample name and just zeros #
        import pandas
        empty_samples = pandas.Series(0, index=self.otus_df.columns)
        # Build an empty new dataframe from a dictionary of empty series #
        result = defaultdict(lambda: empty_samples.copy())
        # Loop over every OTU in the user supplied table #
        for i, otu_counts in self.otus_df.iterrows():
            # The name or id of the current OTU #
            otu_name = otu_counts.name
            # Get the assignment of the current OTU from our classification #
            tax = self.classify.queries_by_id[otu_name].taxonomy
            # By default, the root is at the end #
            tax = list(reversed(tax))
            # Make a string #
            tax_name = '; '.join(tax)
            # Add the current counts to that particular taxonomy #
            result[tax_name] += otu_counts
            # If we have the cumulative option, then propagate up the tree #
            if cumulative:
                for step in range(1, len(tax)):
                    tax_name = '; '.join(tax[0:-step])
                    result[tax_name] += otu_counts
        # Convert to a DataFrame #
        result = pandas.DataFrame(result).T
        # Have the assignment as a separate column and not as an index #
        result = result.reset_index()
        result = result.rename(columns = {'index': 'taxonomy'})
        # Function that takes the length of the taxonomic path and tells
        # us which rank it represents
        rank_names  = self.classify.database.rank_names
        def tax_to_rank(t):
            # There is a problem with too many ranks in some databases
            # requiring this.
            length = len(t.split(';')) - 1
            if length > len(rank_names) - 1: return "Strain"
            else: return rank_names[length]
        # Add the rank column that tells the user if it's a genus or a family #
        ranks = result.taxonomy.apply(tax_to_rank)
        result.insert(loc=0, column='rank', value=ranks)
        # Sort the table by the taxonomy string #
        result = result.sort_values(by=['taxonomy'])
        # Return #
        return result


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script to run the `actualize_database` unittest.
"""

# Built-in modules #
import os

# First party modules #

# Third party modules #

# Internal modules #
from crest4.databases import metadata, ssuome, CrestDatabase

# Constants #
env_val = os.environ.get(CrestDatabase.environ_var)

###############################################################################
def test_download_metadata():
    urls = metadata.db_urls
    assert urls

def test_blast_index():
    db = ssuome.blast_db
    assert db

def test_vsearch_index():
    db = ssuome.vsearch_db
    assert db

###############################################################################
if __name__ == '__main__':
    # Some debug information #
    print("Default directory: ",         CrestDatabase.default_dir)
    print("Environment variable: ",      CrestDatabase.environ_var)
    print("Environment value: ",         env_val)
    print("ssuome path: ",               ssuome.path)
    print("ssuome exists: ",             ssuome.path.exists)
    print("ssuome downloaded: ",         ssuome.downloaded)
    print("ssuome URL: ",                ssuome.url)
    # Run the three tests #
    test_download_metadata()
    test_blast_index()
    test_vsearch_index()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
GNUv3 Licensed.
Contact at www.sinclair.bio
"""

# First party modules #
from plumbing.cache import property_cached

###############################################################################
class Query:
    """
    Represents a single sequence (for instance an OTU) along with
    all the information pertaining to it, such as its taxonomic assignment.

    Takes care of assigning taxonomy by using the results of the sequence
    similarity search and a phylogenetic tree as a N-ary directed graph.

    The query parameter is an object coming from biopython of type:
    'Bio.SearchIO._model.query.QueryResult'
    """

    def __init__(self, classify, query):
        # A reference to the parent object #
        self.classify = classify
        # An object containing all the hit results for this sequence #
        self.query = query
        # Get the name #
        self.name = self.query.id
        # Shortcut to the database used #
        self.db = self.classify.database
        # Shortcut to the algorithm used #
        self.algo = self.classify.search_algo

    def __repr__(self):
        """A simple representation of this object to avoid memory addresses."""
        return "<%s object on '%s'>" % (self.__class__.__name__, self.name)

    #------------------------------ Properties -------------------------------#
    @property_cached
    def nodes(self):
        """
        This function will return the nodes numbers in the tree for which this
        sequence got at least one hit in a set. For example: {'1494'}
        """
        # Initialize the set that will hold all the nodes numbers we find #
        nodes = set()
        # Check there was at least one hit #
        if len(self.query.hits) == 0: return nodes
        # Get the score of the best hit #
        if self.algo == 'blast':   top_score = self.query.hsps[0].bitscore
        if self.algo == 'vsearch': top_score = self.query.hsps[0].ident_pct/100
        # Check if the score is good enough to proceed further #
        if top_score < self.classify.min_score: return nodes
        # Calculate the score-drop threshold based on the best hit #
        threshold = top_score * self.classify.score_frac
        # Iterate on the hits until falling below a threshold #
        for hsp in self.query.hsps:
            # Stop if the current bitscore is below our threshold #
            if self.algo == 'blast':   score = hsp.bitscore
            if self.algo == 'vsearch': score = hsp.ident_pct/100
            if score < threshold: break
            # Get the name (or ID) of the current hit e.g. 'DQ448783' #
            hit_id = hsp.hit_id
            # Get the corresponding node number in the tree e.g. '1494' #
            node = self.db.acc_to_node.get(hit_id, False)
            # Check that it was found #
            msg = f"The search hit '{hit_id}' was not found in the tree." \
                  f" The database '{self.db.dir_name}' is probably corrupted."
            if node is False: raise LookupError(msg)
            # Add it to the list #
            nodes.add(node)
        # Return #
        return nodes

    @property_cached
    def assigned_node(self):
        """
        This function will return the node in the tree at which the
        sequence was assigned. This could be the root of the tree or any
        other node. For example: <Tree at 0x15194b8d>
        This function can also return `False` when there are no results.
        """
        # If there are no hits #
        if len(self.nodes) == 0: return False
        # If there is only one hit, then get that node in the tree #
        if len(self.nodes) == 1:
            node_num, = self.nodes
            gen = self.db.tree.search_nodes(name=node_num)
            # Sanity check that the node was found #
            node = next(gen, False)
            msg = f"Node {node_num!r} not found in the tree."
            if node is False: raise LookupError(msg)
        # Retrieve the lowest common node if more than one hit #
        else:
            node = self.db.tree.common_ancestor(self.nodes)
        # Calculate the similarity fraction of the best alignment #
        if self.algo == 'blast':
            ident_num = self.query.hsps[0].ident_num
            algn_span = self.query.fragments[0]._aln_span
            similarity = ident_num / algn_span
        if self.algo == 'vsearch':
            similarity = self.query.hsps[0].ident_pct/100
        # Check the minimum similarity criteria for assigning at a given
        # level and proceed in an ascending fashion up
        # the tree until the similarity is satisfactory.
        while True:
            # Check that the similarity filter is activated #
            if not self.classify.min_smlrty: break
            # Check if we already got all the way up to the root #
            if node.is_root: break
            # Get the minimum value associated with this level #
            smlrty_min = float(self.db.node_to_name[node.name][1])
            # Check if we are above that minimum #
            if similarity > smlrty_min: break
            # Otherwise, go up one level for our classification #
            node = node.up
        # Return #
        return node

    def get_tax(self, node):
        """Function to get the taxonomy name of a node"""
        # Sanity check that the node has a name #
        msg = f"Node {node!r} doesn't have an ID associated."
        if not node.name: raise LookupError(msg)
        return self.db.node_to_name[node.name][0]

    @property_cached
    def taxonomy(self):
        """
        This function will return a list containing the assigned taxonomy.

        For instance:
            ['root', 'Main genome', 'Bacteria', 'Bacteria (superkingdom)',
             'Terrabacteria', 'Actinobacteria', 'Actinobacteria (class)',
             'Micrococcales', 'Micrococcaceae']
        """
        # Check if there were no hits #
        if self.assigned_node is False: return ["No hits"]
        # The taxonomic name of the current node #
        name = self.get_tax(self.assigned_node)
        # Traverse the tree up to the root #
        tree_path = self.assigned_node.ancestors()
        # Get the name of every parent along the way #
        return [name] + [self.get_tax(parent) for parent in tree_path]

    @property_cached
    def tax_string(self):
        """
        This function will return a single comma-separated string containing
        the full assigned taxonomy along with the original name of the
        sequence classified.
        """
        # Make a semicolon separated string #
        tax = '; '.join(reversed(self.taxonomy))
        # Add the name of the query to the beginning line #
        return self.name + '\t' + tax + '\n'